#include <cstdio>
#include <cstring>
#include <sstream>

#include <qblib/stream.hh>
#include <qblib/error.hh>

#include <map>

PrintStreamBase& PrintStreamBase::operator<< (const BAS_INT v)
{
    std::stringstream tmp; tmp << v; WriteWithFitCheck(tmp.str());
}
PrintStreamBase& PrintStreamBase::operator<< (const BAS_LNG v)
{
    std::stringstream tmp; tmp << v; WriteWithFitCheck(tmp.str());
}
PrintStreamBase& PrintStreamBase::operator<< (const BAS_SGN v)
{
    std::stringstream tmp; tmp << v; WriteWithFitCheck(tmp.str());
}
PrintStreamBase& PrintStreamBase::operator<< (const BAS_DBL v)
{
    std::stringstream tmp; tmp << v; WriteWithFitCheck(tmp.str());
}
PrintStreamBase& PrintStreamBase::operator<< (const BAS_STR& v)
{
    WriteWithFitCheck(v);
}
PrintStreamBase& PrintStreamBase::operator<< (const char* v)
{
    WriteWithFitCheck(v);
}
PrintStreamBase& PrintStreamBase::operator<< (const char v)
{
    WriteWithFitCheck(&v, 1);
}
void PrintStreamBase::WriteWithFitCheck(const void* buf, unsigned n)
{
    CheckFit(n);
    Write(buf, n);
}
void PrintStreamBase::DoSPC(int count)
{
    if(count < 0) count = 0;
    int xpos = GetXpos();
    unsigned xsize = GetXsize();
    if(xsize) count %= xsize;
    std::string spaces(count, ' ');
    Write(&spaces[0], spaces.size());
}
void PrintStreamBase::DoTAB(int newpos)
{
    --newpos;
    if(newpos < 0) newpos = 0;
    int xpos = GetXpos();
    unsigned xsize = GetXsize();
    if(xsize) newpos %= xsize;
    
    if(newpos < xpos) { xpos=newpos=0; DoCRLF(); }
    if(newpos > xpos) DoSPC(newpos-xpos);
}
void PrintStreamBase::DoNextTAB()
{
    const unsigned ColumnWidth = 14;
    static const char SpaceBuf[ColumnWidth+1] = "              ";
    int xpos = GetXpos();
    int columnpos   = xpos % ColumnWidth;
    int need_spaces = ColumnWidth - columnpos;
    if(CheckFit(need_spaces))
        Write(SpaceBuf, need_spaces);
}
bool PrintStreamBase::CheckFit(int n_columns)
{
    int xsize = GetXsize(); if(!xsize) return true;
    int xpos = GetXpos();
    if(xpos+n_columns < xsize) return true;
    if(xpos > 0) DoCRLF();
    return false;
}

class QbFPstream: public PrintStreamBase
{
    FILE* fp;
public:
    QbFPstream(int fd) : fp(fdopen(fd, "wb")) { }
    
    virtual unsigned GetXpos() const
    {
    }
    virtual unsigned GetXsize() const
    {
    }
    virtual void DoCRLF()
    {
    }
    virtual void Write(const void* buf, unsigned n)
    {
    }
};

class InputImplement: public InputStreamBase
{
protected:
    bool InputError;
public:
    virtual void InputTo(BAS_INT& target)
    {
        IgnoreBlankInput();
        long l = InputLong();
        target = l;
        if(l < -32768 || l > 32767) GotOverflow();
    }

    virtual void InputTo(BAS_SGN& target)
    {
        IgnoreBlankInput();
        target = InputLong();
    }

    virtual void InputTo(BAS_DBL& target)
    {
        IgnoreBlankInput();
        target = InputLong();
    }

    virtual void InputTo(BAS_LNG& target)
    {
        IgnoreBlankInput();
        target = InputLong();
    }

    virtual void InputTo(BAS_STR& target)
    {
        IgnoreBlankInput();
        // anything until comma or eol or eof
        // ignore spaces at beginning and end
        // if quote encountered while string empty, parse until quote or eol or eof.
        //    After quote, ignore spaces.
    }
    
protected:
    virtual int InputChar() = 0;
    virtual void GotOverflow() = 0;

    long double InputLong()
    {
        std::string tmp;
    }
    void IgnoreBlankInput()
    {
    }
};

class QbFIstream: public InputImplement
{
    FILE* fp;
public:
    QbFIstream(int fd) : fp(fdopen(fd, "rb")) { }

    virtual void ReadLine(BAS_STR& target)
    {
    }

    virtual void BeginInput()
    {
        InputError = false;
        // File:
        //   Do nothing. InputTo reads file.
    }

    virtual bool EndInputOkp()
    {
        // File:
        //   Return false if InputTo failed, true otherwise.

        if(InputError) return false;
    }

    virtual BAS_STR Input_N(unsigned n)
    {
    }

    virtual int InputChar()
    {
    }

    virtual void GotOverflow()
    {
        B_ERROR(6); // overflow
        InputError = true;
    }
};

class QbConsoleStream: public PrintStreamBase, public InputImplement
{
    BAS_STR InputLineBuffer;
    unsigned InputLineBufferPos;
public:
    QbConsoleStream() : PrintStreamBase(), InputImplement() { }
    
    virtual unsigned GetXpos() const
    {
    }
    virtual unsigned GetXsize() const
    {
    }
    virtual void DoCRLF()
    {
    }
    virtual void Write(const void* buf, unsigned n)
    {
    }

    virtual void ReadLine(BAS_STR& target)
    {
    }

    virtual void BeginInput()
    {
        InputError = false;
        // Console:
        //   Read line into buffer. InputTo queries buffer.
        //   FIXME: should keep cursor on same line.
        ReadLine(InputLineBuffer);
        InputLineBufferPos=0;
    }

    virtual bool EndInputOkp()
    {
        // Console:
        //   Return false if InputTo failed.
        //   Return false if buffer is not empty.
        //   Return true otherwise.

        if(InputError) return false;
        IgnoreBlankInput();
        if(InputLineBufferPos < InputLineBuffer.size()) return false;
        return true;
    }

    virtual BAS_STR Input_N(unsigned n)
    {
    }

    virtual int InputChar()
    {
        if(InputLineBufferPos < InputLineBuffer.size())
        {
            return InputLineBuffer[InputLineBufferPos++];
        }
        return -1;
    }

    virtual void GotOverflow()
    {
        InputError = true;
        *this << "Overflow\n";
    }
};

struct FileDescriptor
{
    QbFPstream fpout;
    QbFIstream fpin;
};
static std::map<int/*qbasic filenum*/, FileDescriptor> fdmap;

PrintStreamBase& GET_STREAM_OUTPUT(int filenumber)
{
    std::map<int, FileDescriptor>::iterator
        i = fdmap.find(filenumber);

    if(i == fdmap.end())
        B_ERROR(52); // bad file number
    
    return i->second.fpout;
}

InputStreamBase& GET_STREAM_INPUT(int filenumber)
{
    std::map<int, FileDescriptor>::iterator
        i = fdmap.find(filenumber);

    if(i == fdmap.end())
        B_ERROR(52); // bad file number
    
    return i->second.fpin;
}

static QbConsoleStream cons;

PrintStreamBase& GET_STDOUT()
{
    return cons;
}
InputStreamBase& GET_STDIN()
{
    return cons;
}


const BAS_STR B_FuncINKEY_S()
{
}
